"""General page functionality."""
import datetime
import logging
import os

from flask import Blueprint, Markup, abort
from flask import current_app as app
from flask import make_response, redirect, render_template, request
from tzlocal import get_localzone

from incorporealcms.lib import get_meta_str, init_md

logger = logging.getLogger(__name__)

bp = Blueprint('pages', __name__, url_prefix='/')


@bp.route('/', defaults={'path': 'index'})
@bp.route('/<path:path>')
def display_page(path):
    """Get the file contents of the requested path and render the file."""
    try:
        resolved_path = request_path_to_instance_resource_path(path)
        logger.debug("received request for path '%s', resolved to '%s'", path, resolved_path)
    except PermissionError:
        abort(400)
    except IsADirectoryError:
        return redirect(f'{path}/', code=301)
    except FileNotFoundError:
        abort(404)

    try:
        with app.open_instance_resource(resolved_path, 'r') as entry_file:
            parent_navs = generate_parent_navs(path)
            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(entry_file.name), get_localzone())
            entry = entry_file.read()
    except FileNotFoundError:
        logger.warning("requested path '%s' (resolved path '%s') not found!", path, resolved_path)
        abort(404)
    else:
        md = init_md()
        content = Markup(md.convert(entry))
        logger.debug("file metadata: %s", md.Meta)

        return render('base.html', title=get_meta_str(md, 'title'), description=get_meta_str(md, 'description'),
                      image=get_meta_str(md, 'image'), base_url=request.base_url, content=content, navs=parent_navs,
                      mtime=mtime.strftime('%Y-%m-%d %H:%M:%S %Z'))


def render(template_name_or_list, **context):
    """Wrap Flask's render_template.

    * Determine the proper site theme to use in the template and provide it.
    """
    PAGE_STYLES = {
        'dark': 'css/dark.css',
        'light': 'css/light.css',
    }

    selected_style = request.args.get('style', None)
    if selected_style:
        user_style = selected_style
    else:
        user_style = request.cookies.get('user-style')
        logger.debug("user style cookie: %s", user_style)
    context['user_style'] = PAGE_STYLES.get(user_style, PAGE_STYLES.get(app.config['DEFAULT_PAGE_STYLE']))

    resp = make_response(render_template(template_name_or_list, **context))
    if selected_style:
        resp.set_cookie('user-style', selected_style)
    return resp


def request_path_to_instance_resource_path(path):
    """Turn a request URL path to the full page path.

    flask.Flask.open_instance_resource will open a file like /etc/hosts if you tell it to,
    which sucks, so we do a lot of work here to make sure we have a valid request to
    something inside the pages dir.
    """
    # check if the path is allowed
    base_dir = os.path.realpath(f'{app.instance_path}/pages/')
    resolved_path = os.path.realpath(os.path.join(base_dir, path))
    logger.debug("base_dir: %s, constructed resolved_path: %s", base_dir, resolved_path)

    # bail if the requested real path isn't inside the base directory
    if base_dir != os.path.commonpath((base_dir, resolved_path)):
        logger.warning("client tried to request a path '%s' outside of the base_dir!", path)
        raise PermissionError

    # if this is a file-like requset but actually a directory, redirect the user
    if os.path.isdir(resolved_path) and not path.endswith('/'):
        logger.info("client requested a path '%s' that is actually a directory", path)
        raise IsADirectoryError

    # derive the proper markdown file depending on if this is a dir or file
    if os.path.isdir(resolved_path):
        absolute_resource = os.path.join(resolved_path, 'index.md')
    else:
        absolute_resource = f'{resolved_path}.md'

    logger.info("final path = '%s' for request '%s'", absolute_resource, path)

    # does the final file actually exist?
    if not os.path.exists(absolute_resource):
        logger.warning("requested final path '%s' does not exist!", absolute_resource)
        raise FileNotFoundError

    # we checked that the file exists via absolute path, but now we need to give the path relative to instance dir
    return absolute_resource.replace(f'{app.instance_path}{os.path.sep}', '')


def resolve_page_file(path):
    """Manipulate the request path to find appropriate page file.

    * convert dir requests to index files

    Worth noting, Flask already does stuff like convert '/foo/../../../bar' to
    '/bar', so we don't need to take care around file access here.
    """
    if path.endswith('/'):
        path = f'{path}index'
    path = f'pages/{path}.md'
    return path


def generate_parent_navs(path):
    """Create a series of paths/links to navigate up from the given path."""
    # derive additional path/location stuff based on path
    resolved_path = resolve_page_file(path)
    parent_dir = os.path.dirname(resolved_path)
    parent_path = '/'.join(path[:-1].split('/')[:-1]) + '/'

    logger.debug("path: '%s'; parent path: '%s'; resolved path: '%s'; parent dir: '%s'",
                 path, parent_path, resolved_path, parent_dir)

    if path in ('index', '/'):
        return [(app.config['TITLE_SUFFIX'], '/')]
    else:
        with app.open_instance_resource(resolved_path, 'r') as entry_file:
            entry = entry_file.read()
        # for issues regarding parser reuse (see lib.init_md) we reinitialize the parser here
        md = init_md()
        _ = Markup(md.convert(entry))
        page_name = " ".join(md.Meta.get('title')) if md.Meta.get('title') else f'/{path}'
        return generate_parent_navs(parent_path) + [(page_name, f'/{path}')]
